# pylint: disable=broad-exception-caught
"""Utilities."""

from enum import Enum
import os
from pathlib import Path
import platform
import pwd
import socket
from typing import Any, List, Union
import base64
import re

from pytoolkit.static import ENCODING, RE_DOMAIN, RE_IP4, SANATIZE_KEYS
from pytoolkit.utilities import flatten_dictionary, nested_dict


def os_plat() -> str:
    """
    Return OS System.

    :return: darwin, linux, java, windows.
    :rtype: str
    """
    return platform.system().lower()


def verify_list(value: Any) -> List[str]:
    """
    Verify value being passed is a list or split out a comma seperted string into a list.

    :param value: Original value. Should be a str|list
    :type value: Any
    :raises ValueError: If value is not a string or list
    :return: _description_
    :rtype: List[str]
    """
    if not isinstance(value, list) and isinstance(value, str):
        return value.split(",")
    if isinstance(value, list):
        return value  # type: ignore
    raise ValueError(f"Invalid value {value}")


def convert_to_base64(filename: str) -> bytes:
    """
    Convert a file to a byte string off base64.

    :param filename: Filename
    :type filename: str
    :return: Encoded File String.
    :rtype: base64
    """
    with open(filename, "rb") as file:
        my_string: bytes = base64.b64decode(file.read())
    return my_string


# Enumerator type


def enum(*sequential: Any, **named: Any) -> type[Enum]:
    """
    Support for converting the values back to names can be added.

    Usage:
        >>> Numbers = enum(ONE=1, TWO=2, THREE='three')
        >>> Numbers.ONE
        1
        >>> Numbers.TWO
        2
        >>> Numbers.THREE
        'three'

    :return: Enumerated Object.
    :rtype: type[Enum]
    """
    enums = dict(zip(sequential, range(len(sequential))), **named)
    reverse = dict(((v, k) for (k, v) in enums.items()))
    enums["reverse_mapping"] = reverse
    return type("Enum", (), enums)


def isstring(arg: Any) -> bool:
    """Verifies if an argument is a string."""
    try:
        return isinstance(arg, basestring)
    except NameError:
        return isinstance(arg, (str, bytes))


# Convenience methods used internally by module
# Do not use these methods outside the module


def string_or_list(value: Any, delimeters: Union[str, None] = None) -> list[str]:
    """
    Return a list containing value.

    This method allows flexibility in class __init__ arguments,
    allowing you to pass a string, object, list, or tuple.
    In all cases, a list will be returned.

    :param value: a string, object, list, or tuple
    :type value: str|obj|list|tuple
    :param delimeter: use a delimeter in the string using pipe(|) as an OR for multiple.
     (Optional) Default no delimeter used. Example: delimeters=',| |;|' or ',| |\|'
    :type delimeter: str|None
    :return: list
    :rtype: list[str]

    :examples:
        "string" -> [string]
        ("t1", "t2") -> ["t1", "t2"]
        ["l1", "l2"] -> ["l1", "l2"]
        None -> None
    """
    if value is None:
        return None  # type: ignore
    if isstring(value):
        return (
            re.split(delimeters, value, flags=re.IGNORECASE) if delimeters else [value]
        )
    return (
        list(value)
        if "__iter__" in dir(value)
        else [
            value,
        ]
    )


def reformat_exception(error: Exception) -> str:
    """
    Reformates Exception to print out as a string pass for logging.

    :param error: caught excpetion
    :type error: Exception
    :return: error as string
    :rtype: str
    """
    resp: str = f"{type(error).__name__}: {str(error)}" if error else ""
    # Replacing [ ] with list() due to issues with reading that format with some systems.
    resp = re.sub(r"\'", "", resp)
    resp = re.sub(r"\[", "list(", resp)
    resp = re.sub(r"\]", ")", resp)
    return resp


def return_filelines(filename: str) -> list[str]:
    """
    Return list of strings in a file.

    :param filename: _description_
    :type filename: str
    :return: _description_
    :rtype: list[str]
    """
    filelines: list[str] = []
    with open(filename, "r", encoding=ENCODING) as fil:
        filelines = fil.readlines()
    return filelines


def check_file(filename: str) -> str:
    """Check that filename exists and returns Pathlib object if does.

    :param filename: Name of file; full path
    :type filename: str
    :raises FileExistsError: _description_
    :return: File location
    :rtype: Path
    """
    file: Path = Path(filename)
    if not file.exists():
        raise FileExistsError(f"Filename does not exist: {str(filename)}")
    return filename


def return_username(log: Any = None) -> Union[str, None]:
    """
    Return Username Information.

    :param log: logger, defaults to None
    :type log: Logger, optional
    :return: username
    :rtype: Union[str,None]
    """
    try:
        return pwd.getpwuid(os.getuid())[0]
    except Exception as err:
        error: str = reformat_exception(err)
        if log:
            log.error(f'msg="Unable to get username"|{error=}')
    return None


def gethostipaddr(hostname: str) -> str:
    """
    Returns IP address of local host. Caution if multiple addresses are rturne due to load balancer.

    :param hostname: _description_
    :type hostname: str
    :raises ValueError: _description_
    :return: _description_
    :rtype: str
    """
    ipv4 = socket.gethostbyname(hostname)
    if not re.match(RE_IP4, ipv4):
        raise ValueError(f"Invalid Address {ipv4}")
    return f"{ipv4}/32" if ipv4.split("/")[-1] != "32" else ipv4


def gethostbyaddr(ip_addr: str) -> str:
    """
    Return FQDN from IP Address.

    :param ip_addr: _description_
    :type ip_addr: str
    :return: _description_
    :rtype: str
    """
    if not re.match(RE_IP4, ip_addr):
        raise ValueError(f"Invalid IPv4 {ip_addr}")
    return socket.gethostbyaddr(ip_addr)[0]


def return_hostinfo(fqdn: bool = True) -> str:
    """
    Return Hostname information on system.

    :param fqdn: Retun FQDN or Hostname, defaults to True
    :type fqdn: bool, optional
    :return: System Hostname/FQDN or root domain.
    :rtype: str
    """
    if fqdn:
        return socket.getfqdn()
    host: str = socket.gethostname()
    if re.match(RE_DOMAIN, host, re.IGNORECASE):
        return (
            ".".join(host.split(".")[:-2])
            if ".".join(host.split(".")[:-2]) != ""
            else ".".join(host.split(".")[:-1])
        )
    return host


def set_bool(value: str, default: bool = False) -> Union[str, bool]:
    """sets bool value when pulling string from os env

    Args:
        value (str|bool, Required): the value to evaluate
        default (bool): default return bool value. Default False

    Returns:
        (str|bool): String if certificate path is passed otherwise True|False
    """
    value_bool = default
    if isinstance(value, bool):
        value_bool = value
    elif str(value).lower() == "true":
        value_bool = True
    elif str(value).lower() == "false":
        value_bool = False
    elif Path.exists(Path(value)):
        value_bool = value
    return value_bool


def sanatize_data(
    data: dict[str, Any], keys: list[str] = SANATIZE_KEYS
) -> dict[str, Any]:
    """
    Sanatize Data from a dictionary of values if a string is found to mask values that should not be exposed.

    :param data: _description_
    :type data: dict[str,Any]
    :param keys: _description_, defaults to SANATIZE_KEYS
    :type keys: list[str], optional
    :return: _description_
    :rtype: dict[str, Any]
    """
    flat = flatten_dictionary(data)
    new_dict = {
        key: "[MASKED]" if isinstance(key, str) and key.lower() in keys else value
        for key, value in flat.items()
    }
    return nested_dict(new_dict)



"""SSH Remote Comman Executions."""


from typing import Iterable, Mapping, Union, Any
from dataclasses import dataclass

from paramiko import PKey
from paramiko.auth_strategy import AuthStrategy

from pytoolkit.utilities import BaseMonitor
from pytoolkit.static import SSH_PORT


@dataclass
class SubProcessReturn(BaseMonitor):
    """Subprocess return."""

    stdin: Any
    stdout: Any
    stderr: Any
    returncode: int
    command: str
    timeout: int
    get_pty: bool


@dataclass
class Connector(BaseMonitor):
    """Paramiko Connector Variables"""

    hostname: str
    port: int = SSH_PORT
    username: Union[str, None] = None
    password: Union[str, None] = None
    pkey: Union[PKey, None] = None
    key_filename: Union[str, None] = None
    timeout: Union[float, None] = None
    allow_agent: bool = True
    look_for_keys: bool = True
    compress: bool = False
    sock: Any = None
    gss_auth: bool = False
    gss_kex: bool = False
    gss_deleg_creds: bool = True
    gss_host: Union[str, None] = None
    banner_timeout: Union[float, None] = None
    auth_timeout: Union[float, None] = None
    channel_timeout: Union[float, None] = None
    gss_trust_dns: bool = True
    passphrase: Union[str, None] = None
    disabled_algorithms: Union[Mapping[str, Iterable[str]], None] = None
    transport_factory: Any = None
    auth_strategy: Union[AuthStrategy, None] = None



"""Splunk Integrations."""

from collections import OrderedDict
import datetime
from typing import Any, Optional, Union

from dataclasses import dataclass
import requests

from pytoolkit.utilities import BaseMonitor, NONETYPE

splunk_url = "{}//{}:{}"

@dataclass
class SplunkHecHeader(BaseMonitor):
    """Splunk Hec Header."""

    splunk_server: str
    token: str
    sourcetype: Optional[str] = NONETYPE
    source: Optional[str] = NONETYPE
    index: Optional[str] = NONETYPE
    splunk_port: int = 8088
    verify: Union[str, bool] = True
    upload: bool = True
    timeout: float = 15.0
    schema: str = "https"


def splunk_format(**kwargs: Any) -> str:
    """
    Reformat a list of key:value pairs into a simple logging message for Splunk.

    :return: _description_
    :rtype: str
    """
    ordered: OrderedDict[str, Any] = OrderedDict(sorted(kwargs.items()))
    string: list[str] = [f'{str(key)}="{value}"' for key, value in ordered.items()]
    return ",".join(string)


# TODO: fix the ability to send HEC formats in chuncks. ippoad
def splunk_hec_format(
    host: str,
    source: str,
    sourcetype: str,
    metrics_list: Union[list[str], None] = None,
    **kwargs: Any,
) -> dict[str, Any]:
    """Create a JSON style hec format.

    :param host: hostname
    :type host: str
    :param source_name: source of dataa
    :type source_name: str
    :param metrics_list: list of metrics type fileds found in arguments
    :type metrics_list: list[str]
    :param kwargs: key:value pairs to extract and format data structure
    :return: Splunk Hec Datastructure
    :rtype: dict[str,Any]
    """
    hec_json: dict[str, Any] = {
        "time": kwargs.pop("time", datetime.datetime.now().timestamp()),
        "host": host,
        "source": source,
        "sourcetype": sourcetype,
        "event": {},
    }
    if metrics_list:
        # Build HEC style Metrics
        hec_json["fields"] = {
            f"metric_name:{metric}": kwargs.pop(metric, None) for metric in metrics_list
        }
        hec_json["fields"] = dict(sorted(hec_json["fields"].items()))
    hec_json["events"] = {**hec_json["events"], **kwargs}
    hec_json["events"] = dict(sorted(hec_json["events"].items()))
    return hec_json

def splunk_upload(
    server: str,
    token: str,
    hec_data: Any,
    verify: Any,
    timeout: float,
    port: int = 8088,
    chunk_size: int = 0,
):
    """
    Upload Splunk Data.

    :param server: _description_
    :type server: str
    :param token: _description_
    :type token: str
    :param hec_data: _description_
    :type hec_data: Any
    :param verify: _description_
    :type verify: str
    :param port: _description_, defaults to 8088
    :type port: int, optional
    :param chunk_size: Set size to split up data into if too large;
     hec_data must be a list of json entries, defaults to 0 or all data
    :type chunk_size: int, optional
    """
    EVENTPATH = "services/collector/event"
    # split_list = list(split(hec_json_list, 10))
    # for server in server_list:
    url = f"https://{server}:{port}/{EVENTPATH}"
    headers = {
        "Content-Type": "application/json",
        "Authorization": f"Splunk {token}",
        "X-Splunk-Request-Channel": token,
    }
    payload = hec_data
    response = requests.post(
        url, headers=headers, json=payload, verify=verify, timeout=timeout
    )
    msg = f'msg="uploaded splunk data response"|status_code={response.status_code}, response={response.json()}'
    response.raise_for_status()



# pylint: disable=invalid-name
"""Files."""

import re
import json
from typing import Any, Union
from pathlib import Path
import platform
import tempfile

import yaml

from pytoolkit.static import ENCODING, FILE_UMASK_PERMISSIONS, CONFIG_PATH


class BytesDump(json.JSONEncoder):
    """Resovlve error with byte present in Dict."""

    def default(self, o: Any) -> Union[str, Any]:
        """Resolve error with byte in Dict."""
        if isinstance(o, bytes):
            return o.decode()
        return json.JSONEncoder.default(self, o)


def get_config_section(filename: str, ftype: str, section: str = ""):
    """
    Returns a configuration dictionary or a section of a configuration if found.

    :param filename: _description_
    :type filename: str
    :param ftype: _description_
    :type ftype: str
    :param section: _description_, defaults to ""
    :type section: str, optional
    :return: _description_
    :rtype: _type_
    """
    settings = read_yaml(Path(filename))
    if section:
        settings = settings.get(section, {})
    return settings


def read_yaml(filename: Path) -> dict[str, Any]:
    """
    Read in a YAML configuration file.

    :param filename: Yaml File Full Path
    :type filename: Path
    :return: Yaml Configurations
    :rtype: dict[str,Any]
    """
    check_file(filename=str(filename))
    with open(filename, "r", encoding=ENCODING) as r_yaml:
        settings: Any = yaml.safe_load(r_yaml)
    return settings


def get_tempdir() -> str:
    """Returns tempdir"""
    return tempfile.gettempdir()


def get_var_dir(extend_path: Union[str, None] = None, mode: str = "default") -> str:
    """
    Get default var directory depending on OS. Extend to application path if path supplied.
     Allows for user to create a directory strcture where reports or logs can be stored. Using
     a variable directory or extend a path in a users home directory.

    :param extend_path: Extends the system default location;
     creates a new app directory, defaults to None
    :type extend_path: str, optional
    :return: Log Directory for Logs
    :rtype: str
    """
    directory: dict[str, Path] = {
        "darwin": Path.joinpath(Path.home() / "Library/Logs"),
        "linux": Path("/var/log"),
    }
    plat: str = platform.system()
    try:
        path = (
            Path(f"{str(directory[plat.lower()])}/{extend_path}")
            if extend_path
            else Path(str(directory[plat.lower()]))
        )
        mkdir(path=path, mode=mode)
        return str(path)
    except KeyError:
        path = (
            Path(f"{str(tempfile.gettempdir())}/{extend_path}")
            if extend_path
            else Path(tempfile.gettempdir())
        )
        mkdir(path=path, mode=mode)
        return str(path)


def set_location(
    location: str, extend_path: Union[str, None] = None, mode: str = "default"
) -> str:
    """Set default logDir or configuration directory based on mode defined."""
    loc = "var"
    if bool(re.match(r"(home|homedir)", location)):
        loc = "home"
    mode = mode if mode in FILE_UMASK_PERMISSIONS else "default"
    base_dir = {"home": set_homedir, "var": get_var_dir}
    return base_dir[loc](extend_path=extend_path, mode=mode)


def set_homedir(extend_path: Union[str, None] = None, mode: str = "default") -> str:
    """
    Return the users home dir and can extend if using a subdirctory.
     Use `mode` to restrict permissions.

    :param extend_path: _description_, defaults to None
    :type extend_path: Union[str,None], optional
    :param mode: _description_, defaults to "default"
    :type mode: str, optional
    :return: _description_
    :rtype: str
    """
    path = Path(Path.home() / extend_path) if extend_path else Path.home()
    mkdir(path=path, mode=mode)
    return str(path)


def mkdir(path: Path, mode: str = "default") -> None:
    """Makes Dir based on permissions passed."""
    # for parent in reversed(path.parents):
    path.mkdir(mode=FILE_UMASK_PERMISSIONS[mode], parents=True, exist_ok=True)


def get_home() -> str:
    """Return current users `home` direcotry."""
    return str(Path.home())


def with_suffix(logName: str) -> str:
    """Add suffix to logname."""
    return str(Path(logName).with_suffix(".log"))


def check_file(filename: str) -> None:
    """
    Checks if a file exists.

    :param filename: _description_
    :type filename: str
    :raises ValueError: _description_
    """
    if not Path(filename).is_file():
        raise ValueError(f"Not a file {filename}")


def get_config_location(
    config_location: list[str],
    app_name: Union[str, None] = None,
    file_format: str = "yml",
) -> str:
    """
    Retrieve configuraiton lcoation if one exists in the paths to search.

        Ex:
            config_location = [
                str(Path.joinpath(Path.home() / ".config/application.yaml")),
                str(Path.joinpath(Path.home() / ".config/application.yml")),
                str(Path("/etc/appname/appconfig.yaml")),
                str(Path("/Users/guest/Library/logs/appname/appconfig.yml"))
            ]
        OR:
            config_location = [
                str(Path.joinpath(Path.home() / ".config/application.yaml")),
                str(Path.joinpath(Path.home() / ".config/application.yml")),
                str(Path("/etc/appname/appconfig.yaml")),
                str(Path("/Users/guest/Library/logs/appname/appconfig.yml"))
            ]
    :param config_location: _description_
    :type config_location: list[str]
    :return: _description_
    :rtype: str
    """
    for location in config_location:
        # TODO: Allow for diff types as currentlu just allowing yaml.
        # if Path.is_file(Path(
        #    CONFIG_PATH.format(location,app_name, file_format) if app_name else location)
        # ):
        if Path.is_file(Path(location)):
            return location
        return ""



# pylint: disable=line-too-long
"""Global Static Vars."""

from typing import cast

ENCODING: str = "utf-8"

# See https://www.linuxtrainingacademy.com/all-umasks/
FILE_UMASK_PERMISSIONS = {
    "default": 0o40775,  # umask 002 rw-rw-r- rwxrwx-r
    "restrictive": 0o40770,  # umask 037 rw-r--- rwxr---
    "root": 0o40755,  # umask 022 rw-r-r- rwxr-xr-x
}

NONETYPE: None = cast(None, object())
DEFAULT_TO: list[str] = ["john.doe@acme.com"]
DEFAULT_FROM: str = "python-script@acme.com"
DEFAULT_CC: list[str] = [""]
DEFAULT_BCC: list[str] = [""]
TMP_PEM_POSTFIX = "_pytookit.pem"
PEM_REGEX = (
    r".*[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}_pytoolkit\.pem$"
)

RE_DOMAIN = r".*\.(com|net|gov)$"
RE_IP4 = r"^((25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])\.){3}(25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])$"
SANATIZE_KEYS = [
    "password",
    "cacontext",
    "log",
    "plaintext",
    "token",
    "jwt",
    "cert",
    "authorization",
    "auth",
]
CONFIG_PATH = "{}/{}.{}"
SSH_PORT = 22
DISABLED_ALGORITHMS = {
    "ciphers": [
        "aes128-cbc",
        "aes192-cbc",
        "aes256-cbc",
        "3des-cbc",
    ],
    "macs": ["hmac-sha1", "hmac-md5", "hmac-sha1-96", "hmac-md5-96"],
    "keys": ["ssh-dss", "ssh-rsa"],
    "pubkeys": ["ssh-rsa", "ssh-dss"],
    "kex": [
        "diffie-hellman-group1-sha1",
        "diffie-hellman-group14-sha1",
        "diffie-hellman-group-exchange-sha1",
    ],
}



"""SSH Remote Comman Executions."""
import subprocess
from typing import Union, Any

import paramiko
from paramiko import AutoAddPolicy
from paramiko.client import SSHClient

from pytoolkit.ops.models import SubProcessReturn, Connector
from pytoolkit.static import DISABLED_ALGORITHMS, SSH_PORT
from pytoolkit.utils import string_or_list

def _ssh_client():
    return paramiko.SSHClient()


class SSHConnector:
    _missing_host_key_policy = AutoAddPolicy()
    _private_key = ""
    _timeout: float = 120.0
    _disabled_algorithms = DISABLED_ALGORITHMS

    def __init__(
        self,
        hostname: str,
        username: Union[str, None] = None,
        password: Union[str, None] = None,
        port: int = SSH_PORT,
        **kwargs: Any,
    ):
        self.client: SSHClient = _ssh_client()
        self.hostname = hostname
        self._timeout = kwargs.pop("timeout", self._timeout)
        banner_timeout = kwargs.pop("banner_timeout", self._timeout)
        auth_timeout = kwargs.pop("auth_timeout", self._timeout)
        disabled_algorithms = kwargs.pop(
            "disabled_algorithms", self._disabled_algorithms
        )
        if kwargs.get("auto_add", False) == True:
            self.client.set_missing_host_key_policy(self._missing_host_key_policy)
        self.connection_settings = Connector.create_from_kwargs(
            **{
                **{
                    "hostname": hostname,
                    "username": username,
                    "password": password,
                    "port": port,
                    "timeout": self._timeout,
                    "banner_timeout": banner_timeout,
                    "auth_timeout": auth_timeout,
                    "disabled_algorithms": disabled_algorithms,
                },
                **kwargs,
            }
        )
        self.sftp = None

    @property
    def disabled_algorithms(self):
        return self._disabled_algorithms

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, value: float):
        self._timeout = value

    def connect(self):
        """
        SSHClient Context Manager.

        Connect to an SSH server and authenticate to it.  The server's host key
        is checked against the system host keys (see `load_system_host_keys`)
        and any local host keys (`load_host_keys`).  If the server's hostname
        is not found in either set of host keys, the missing host key policy
        is used (see `set_missing_host_key_policy`).  The default policy is
        to reject the key and raise an `.SSHException`.

        Authentication is attempted in the following order of priority:

            - The ``pkey`` or ``key_filename`` passed in (if any)

              - ``key_filename`` may contain OpenSSH public certificate paths
                as well as regular private-key paths; when files ending in
                ``-cert.pub`` are found, they are assumed to match a private
                key, and both components will be loaded. (The private key
                itself does *not* need to be listed in ``key_filename`` for
                this to occur - *just* the certificate.)

            - Any key we can find through an SSH agent
            - Any "id_rsa", "id_dsa" or "id_ecdsa" key discoverable in
              ``~/.ssh/``

              - When OpenSSH-style public certificates exist that match an
                existing such private key (so e.g. one has ``id_rsa`` and
                ``id_rsa-cert.pub``) the certificate will be loaded alongside
                the private key and used for authentication.

            - Plain username/password auth, if a password was given

        If a private key requires a password to unlock it, and a password is
        passed in, that password will be used to attempt to unlock the key.

        :param str hostname: the server to connect to
        :param int port: the server port to connect to
        :param str username:
            the username to authenticate as (defaults to the current local
            username)
        :param str password:
            Used for password authentication; is also used for private key
            decryption if ``passphrase`` is not given.
        :param str passphrase:
            Used for decrypting private keys.
        :param .PKey pkey: an optional private key to use for authentication
        :param str key_filename:
            the filename, or list of filenames, of optional private key(s)
            and/or certs to try for authentication
        :param float timeout:
            an optional timeout (in seconds) for the TCP connect
        :param bool allow_agent:
            set to False to disable connecting to the SSH agent
        :param bool look_for_keys:
            set to False to disable searching for discoverable private key
            files in ``~/.ssh/``
        :param bool compress: set to True to turn on compression
        :param socket sock:
            an open socket or socket-like object (such as a `.Channel`) to use
            for communication to the target host
        :param bool gss_auth:
            ``True`` if you want to use GSS-API authentication
        :param bool gss_kex:
            Perform GSS-API Key Exchange and user authentication
        :param bool gss_deleg_creds: Delegate GSS-API client credentials or not
        :param str gss_host:
            The targets name in the kerberos database. default: hostname
        :param bool gss_trust_dns:
            Indicates whether or not the DNS is trusted to securely
            canonicalize the name of the host being connected to (default
            ``True``).
        :param float banner_timeout: an optional timeout (in seconds) to wait
            for the SSH banner to be presented.
        :param float auth_timeout: an optional timeout (in seconds) to wait for
            an authentication response.
        :param float channel_timeout: an optional timeout (in seconds) to wait
             for a channel open response.
        :param dict disabled_algorithms:
            an optional dict passed directly to `.Transport` and its keyword
            argument of the same name.
        :param transport_factory:
            an optional callable which is handed a subset of the constructor
            arguments (primarily those related to the socket, GSS
            functionality, and algorithm selection) and generates a
            `.Transport` instance to be used by this client. Defaults to
            `.Transport.__init__`.
        :param auth_strategy:
            an optional instance of `.AuthStrategy`, triggering use of this
            newer authentication mechanism instead of SSHClient's legacy auth
            method.

            .. warning::
                This parameter is **incompatible** with all other
                authentication-related parameters (such as, but not limited to,
                ``password``, ``key_filename`` and ``allow_agent``) and will
                trigger an exception if given alongside them.

        :returns:
            `.AuthResult` if ``auth_strategy`` is non-``None``; otherwise,
            returns ``None``.

        :raises BadHostKeyException:
            if the server's host key could not be verified.
        :raises AuthenticationException:
            if authentication failed.
        :raises UnableToAuthenticate:
            if authentication failed (when ``auth_strategy`` is non-``None``;
            and note that this is a subclass of ``AuthenticationException``).
        :raises socket.error:
            if a socket error (other than connection-refused or
            host-unreachable) occurred while connecting.
        :raises NoValidConnectionsError:
            if all valid connection targets for the requested hostname (eg IPv4
            and IPv6) yielded connection-refused or host-unreachable socket
            errors.
        :raises SSHException:
            if there was any other error connecting or establishing an SSH
            session.
        """
        self.client.connect(**self.connection_settings.to_dict())

    def exec_command(self, **kwargs: Any) -> SubProcessReturn:
        """
        (method) def exec_command(
        command: str,
        bufsize: int = -1,
        timeout: Union[float, None] = None,
        get_pty: bool = False,
        environment: Mapping[str, str] | None = None
        )
        """
        timeout = kwargs.pop("timeout", self._timeout)
        get_pty = kwargs.pop("get_pty", False)
        # try:
        self.connect()
        stdin, stdout, stderr = self.client.exec_command(
            timeout=timeout, get_pty=get_pty, **kwargs
        )
        return SubProcessReturn(
            stderr=stderr,
            stdin=stdin,
            stdout=stdout,
            returncode=255 if not stderr.readline() else 0,
            timeout=timeout,
            get_pty=get_pty,
            command=kwargs["command"],
        )
        # except paramiko.SSHException as err:
        # return SubProcessReturn(stderr=stderr,stdin=stdin,stdout=stdout,returncode=255 if not stderr.readline() else 0,timeout=timeout,get_pty=get_pty,command=kwargs["command"])

    def open_sftp(self):
        self.sftp = self.client.open_sftp()


def return_subprocess_error(**kwargs: Any) -> subprocess.CompletedProcess[Any]:
    """
    Reformates exception for subprocess error.

    :return: _description_
    :rtype: subprocess.CompletedProcess[Any]
    """
    error = kwargs["error"]
    return subprocess.CompletedProcess(args=[], stdout="", returncode=255, stderr=error)


def exec_cmd(
    host: str,
    command: str,
    user: Union[str, None] = None,
    key: Union[str, None] = None,
    check: bool = False,
    text: bool = True,
    capture_output: bool = True,
    timeout: float = 5.0,
) -> subprocess.CompletedProcess[Any]:
    """
    Runs Command via subprocess. Configure local ssh first if using defaults if executing remote command.

    :param host: target host to send the command to
    :type host: str
    :param command: command to run on the host
    :type command: str
    :param user: user to use to login to host, defaults to None
    :type user: _type_, optional
    :param key: _description_, defaults to None
    :type key: _type_, optional
    :param stdin: override sys.stdin, defaults to None
    :type stdin: _type_, optional
    :param check: pass to *subprocess.run*; if set, checks return code
        and raises subprocess.CalledProcessError, if none-zero result, defaults to False
    :type check: bool, optional
    :return: _description_
    :rtype: _type_
    """
    cmd: list[str] = []
    cmd: list[str] = string_or_list(command, delimeters=" |,")  # type: ignore
    exect = []
    if host not in ["self", "localhost"]:
        exect = [host]
    if key:
        exect.extend(["-i", key])
    if user:
        exect.append(user)
    result = subprocess.run(
        exect + cmd,
        shell=False,
        capture_output=capture_output,
        text=text,
        check=check,
        timeout=timeout,
    )
    return result



"""Create a custo CA."""

import re
import tempfile
import uuid

from pathlib import Path
import certifi

from pytoolkit.static import ENCODING, PEM_REGEX, TMP_PEM_POSTFIX
from pytoolkit.utils import check_file, return_filelines, string_or_list


def create_custom_cert(certstore: list[str]) -> tuple[str, list[str]]:
    """
    Creates a temp custom certificate.

    :param certstore: list of pem certificates to use.
    :type certstore: list[str]
    :return: Returns CA_TMP_FILE, Raw CA_CONTEXT
    :rtype: tuple[str,list[str]]
    """
    # verify correct type
    certstore = string_or_list(value=certstore)
    # create context
    ca_context = castore_create_context(certstore=certstore)
    ca_tmp_file = castore_create_tmp(ca_context=ca_context)
    return ca_tmp_file, ca_context


def castore_create_tmp(ca_context: list[str]) -> str:
    """
    Create Temporary file for passing to a high level request.

    :param ca_context: _description_
    :type ca_context: list[str]
    :return: _description_
    :rtype: str
    """
    temp: str = tempfile.gettempdir()
    file_name: str = f"{uuid.uuid4()}{TMP_PEM_POSTFIX}"
    writeout: Path = Path.joinpath(Path(temp) / file_name)
    with open(writeout, "w", encoding=ENCODING) as fil:
        fil.writelines(ca_context)
    return str(writeout)


def castore_create_context(certstore: list[str]) -> list[str]:
    """
    Create a custom CA Store Context for verification.

    :param certstore: List of CA PEM files to add
    :type certstore: list[str]
    :return: Raw Pem file context
    :rtype: list[str]
    """
    castore: str = certifi.where()
    # Create CA context only return as raw list
    castore_context: list[str] = return_filelines(castore)
    if not certstore:
        return castore_context
    for cert in certstore:
        try:
            cert = check_file(filename=cert)
            castore_context.extend(return_filelines(filename=cert))
        except FileExistsError:
            pass
    return castore_context


def castore_custom_delete(custom_castore_loc: str) -> None:
    """
    Deletes Temporary custom CA Created.

    :param custom_castore_loc: Location of temp cert
    :type custom_castore_loc: str
    """
    if custom_castore_loc:
        try:
            castore = check_file(custom_castore_loc)
            # Delete using Pathlib returned value if it is a temp SNC CA
            if re.match(PEM_REGEX, str(castore), re.MULTILINE):
                Path(castore).unlink()
        except FileExistsError:
            # Nothing to delete moving on
            pass



"""Init."""

import functools

from pytoolkit._version import __version__

try:
    from decorator import decorator
except ImportError:

    def decorator(caller):
        """Turns caller into a decorator.
        Unlike decorator module, function signature is not preserved.

        :param caller: caller(f, *args, **kwargs)
        """

        def decor(f):
            @functools.wraps(f)
            def wrapper(*args, **kwargs):
                return caller(f, *args, **kwargs)

            return wrapper

        return decor



"""Data Module."""

from typing import Any
from dataclasses import dataclass, field


sample_headers: dict[str, int] = {"test": 1}


@dataclass
class Response:
    """Mock Response."""

    body: str
    status: int = field(default=200)
    headers: dict[Any, Any] = field(default_factory=dict)

    def __init__(self, body: str, status: int = 200):
        self.body = body
        self.status = status
        self.headers = sample_headers



# pylint: disable=dangerous-default-value
"""Mailer."""

from typing import List, Union
import smtplib
import re

from email import encoders
from email.mime.multipart import MIMEMultipart
from email.mime.base import MIMEBase
from email.mime.text import MIMEText

from pytoolkit.utils import string_or_list
from pytoolkit.static import DEFAULT_TO, DEFAULT_FROM, DEFAULT_CC, DEFAULT_BCC, ENCODING


def send_mail(
    smtp_server: str,
    msg: str = "EMPTY",
    subject: str = "Python Script",
    mail_to: Union[list[str], str] = DEFAULT_TO,
    mail_from: str = DEFAULT_FROM,
    mail_cc: Union[list[str], str] = DEFAULT_CC,
    mail_bcc: Union[list[str], str] = DEFAULT_BCC,
    msg_html: Union[str, None] = None,
    attachment: Union[str, None] = None,
    port: int = 25,
) -> str:
    """Send Mail

    :param smtp_server: _description_
    :type smtp_server: str
    :param msg: _description_, defaults to "EMPTY"
    :type msg: str, optional
    :param subject: _description_, defaults to "Python Script"
    :type subject: str, optional
    :param mail_to: _description_, defaults to _default_to
    :type mail_to: list, optional
    :param mail_from: _description_, defaults to _default_from
    :type mail_from: str, optional
    :param mail_cc: _description_, defaults to _default_cc
    :type mail_cc: list, optional
    :param mail_bcc: _description_, defaults to _default_bcc
    :type mail_bcc: list, optional
    :param msg_html: _description_, defaults to None
    :type msg_html: str, optional
    :param attachment: _description_, defaults to None
    :type attachment: str, optional
    :param port: _description_, defaults to 25
    :type port: int, optional
    :return: _description_
    :rtype: _type_
    """
    # Handle if a string is passed
    delimeters = ",| "
    mail_to: List[str] = string_or_list(value=mail_to, delimeters=delimeters)
    mail_cc: List[str] = string_or_list(value=mail_cc, delimeters=delimeters)
    mail_bcc: List[str] = string_or_list(value=mail_bcc, delimeters=delimeters)

    message = MIMEMultipart("alternative")
    message["Subject"] = subject
    message["From"] = mail_from
    message["To"] = ", ".join(mail_to)
    message["cc"] = ", ".join(mail_cc)
    message["bcc"] = ", ".join(mail_bcc)

    try:
        if attachment and isinstance(attachment, list):
            for attach in attachment:
                # add file as application/octet-stream
                # email client can usually downlaoad this automatically as an attachemment
                part = MIMEBase("application", "octet-stream")
                with open(attach, "rb", encoding=ENCODING) as a:
                    # Add encoded stream to MIMEBase app payload
                    part.set_payload(a.read())
                    # encode
                    encoders.encode_base64(part)
                # Get Filename
                split: list[str] = re.findall(r"[\w']+", attach)
                filename: str = f"{split[-2]}.{split[-1]}"
                # Add header as key/value pair to attach part
                part.add_header(
                    "Content-Disposition",
                    f"attachment; filename={filename}",
                )
                # Add attachment to message and convert message to string
                message.attach(part)
        if msg_html:
            html_email_message = MIMEText(msg_html, "html")
            message.attach(html_email_message)
        else:
            plaintext_email_message = MIMEText(msg, "plain")
            message.attach(plaintext_email_message)
        # Send Email
        with smtplib.SMTP(smtp_server, port) as server:
            server.ehlo()
            server.send_message(message)
            server.quit()
        return "Success"
    except (
        smtplib.SMTPConnectError,
        smtplib.SMTPDataError,
        smtplib.SMTPAuthenticationError,
        smtplib.SMTPHeloError,
        smtplib.SMTPServerDisconnected,
    ) as err:
        return f"SMTP Connection Error: {str(err)}"
    except (
        smtplib.SMTPException,
        smtplib.SMTPSenderRefused,
        smtplib.SMTPNotSupportedError,
        smtplib.SMTPResponseException,
        smtplib.SMTPRecipientsRefused,
    ) as err:
        return f"SMTP Communication Error: {str(err)}"
    except Exception:
        return "SMTP Unknown Error"


# TODO: Create an encoded function to allow encoded/embeded attachements
# TODO: Add basic templates that can be used for HTML notifications based off Jinja2 examples



# pylint: disable=line-too-long
"""Package Supplied Utilities."""

from typing import Any, Callable, Generator, Hashable, List, Union
from collections.abc import MutableMapping

from pathlib import Path
from dataclasses import dataclass, fields, field, is_dataclass
import pandas as pd

from pytoolkit.static import NONETYPE


@dataclass
class BaseMonitor:
    """Base Dataclass Methods."""

    @classmethod
    def create_from_dict(cls, _dict: dict[str, Any]):
        """
        Class Method that returns dataclass using a
         dictionary and strips invalid params.

        :param _dict: Dictionary of Values
        :type _dict: dict[str, Any]
        :return: Dataclass
        :rtype: :dataclass: DataModel
        """
        class_fields: set[str] = {f.name for f in fields(cls)}
        return cls(**{k: v for k, v in _dict.items() if k in class_fields})

    @classmethod
    def create_from_kwargs(cls, **kwargs: Any):
        """
        Class method that returns dataclass values by
          unpacking paramter values and strips out invalid params.

        :param kwargs: unpacked key:value pairs
        :type **kwargs: variable length argument list
        :return: Dataclass
        :rtype: :dataclass: DataModel
        """
        class_fields: set[str] = {f.name for f in fields(cls)}
        return cls(**{k: v for k, v in kwargs.items() if k in class_fields})

    def to_dict(self, extend: bool = True) -> dict[str, Any]:
        """
        Returns dataclass as dictionary.

        :param extend: Extends the dataclass that have a value `NONETYPE`.
        :type bool: Optional True
        :return: dataclass dictionary
        :rtype: dict[str, Any]
        """
        if not extend:
            return dict(self.__dict__.items())
        return {k: v for k, v in self.__dict__.items() if v is not NONETYPE}


def _flatten_dict_gen(
    _d: MutableMapping[str, Any],
    parent_key: str,
    sep: str,
    extended_label: bool,
    skip_item: list[str],
) -> Generator[tuple[str, Any], Any, None]:
    for k, v in _d.items():
        new_key: str = k
        if extended_label:
            new_key: str = (
                parent_key + sep + k if (parent_key and k not in skip_item) else k
            )  # type: ignore
        if isinstance(v, MutableMapping):
            yield from flatten_dict(
                v,  # type: ignore
                new_key,
                extended_label=extended_label,
                skip_item=skip_item,
                sep=sep,
            ).items()
        else:
            yield new_key, v


def flatten_dict(
    _dict: MutableMapping[str, Any],
    parent_key: str = "",
    sep: str = ".",
    extended_label: bool = True,
    skip_item: Union[list[str], None] = None,
) -> dict[str, Any]:
    """
    Flatten out a dictionary with nested values.

    :param _dict: Dictionary
    :type _dict: MutableMapping[str,Any]
    :param parent_key: Top Level Key, defaults to ""
    :type parent_key: str, optional
    :param sep: Seperator, defaults to "."
    :type sep: str, optional
    :param extended_label: Uses the same key by default or appends the hierarchy
     into the name of the key used to express the nesting structure, defaults to True
    :type extended_label: bool, optional
    :param skip_item: List of Keys to ignore and flatten without the parent, defaults to []
    :type skip_item: list, optional
    :return: Flattened Dictionary
    :rtype: dict[str,Any]
    """
    skip_item = skip_item if skip_item else [""]
    return dict(_flatten_dict_gen(_dict, parent_key, sep, extended_label, skip_item))


def flatten_dictionary(
    _dict: MutableMapping[Any, Any], sep: str = "."
) -> dict[Hashable, Any]:
    """
    Flatten a dictionary via pandas normalizer.

    :param d: _description_
    :type d: MutableMapping
    :param sep: Adds seperator to Key, defaults to '.'
    :type sep: str, optional
    :return: Flattened Dictionary.
    :rtype: MutableMapping
    """
    [flat_dict] = pd.json_normalize(_dict, sep=sep).to_dict(  # type: ignore
        orient="records"
    )  # type: ignore
    return flat_dict


def _nest_dict_rec(key: str, value: Any, sep: str, out: dict[str, Any]) -> None:
    """Maniplates dictionary into a nested dictionary based on seperator."""
    key, *rest = key.split(sep, 1)
    if rest:
        _nest_dict_rec(rest[0], value, sep, out.setdefault(key, {}))
    else:
        out[key] = value


def nested_dict(_dict: MutableMapping[str, Any], sep: str = ".") -> dict[str, Any]:
    """
    Transform a Flattened Dictionary into a Nested Dictionary.

    :param _dict: _description_
    :type _dict: MutableMapping[str, Any]
    :param sep: Seperator to split Keys with, defaults to '.'
    :type sep: str, optional
    :return: Nested Dictionary.
    :rtype: dict[str, Any]
    """
    # TODO: fix the nested structure add abiltiy to read in a csv or XCEL to help maniplate proper csv human readable datastructures
    result: dict[str, Any] = {}
    for k, v in _dict.items():
        _nest_dict_rec(k, v, sep, result)
    return result


@dataclass
class Matches:
    """Extract a list of matcehs and non matching lists."""

    matches: List[Any] = field(default_factory=lambda: [])
    no_match: List[Any] = field(default_factory=lambda: [])


def set_bool(value: Union[str, bool], default: bool = False) -> Union[str, bool]:
    """
    Sets bool value when pulling string from os env.

    :param value: The value to evaluate
    :type value: str
    :param default: default return bool value, defaults to False
    :type default: bool, optional
    :return: String if a path is passed otherwise True|False
    :rtype: Union[str,bool]
    """
    value_bool: Union[bool, str] = default
    if isinstance(value, bool):
        value_bool = value
    elif str(value).lower() in ["true", "t", "1", "yes", "y"]:
        value_bool = True
    elif str(value).lower() in ["false", "f", "0", "no", "n"]:
        value_bool = False
    elif Path.exists(Path(str(value))):
        value_bool = value
    return value_bool


def extract_matches(
    iterable: Union[list[Any], None], condition: Callable[[list[Any]], Any]
) -> Matches:
    """
    Returns two lists; one that matches the condition and other that does not.
     Use the condition variable to send callable functions used in a regular expression match.

    :param iterable: Lists of Strings.
    :type iterable: Union[list[Any], None]
    :param condition: Callable function or lambda function.
    :type condition: Callable[[list[Any]], Any]
    :return: matches
    :rtype: Matches
    """
    res = Matches([], [])
    if not iterable:
        return res
    return Matches(
        matches=[item for item in iterable if any(condition(item))],
        no_match=[item for item in iterable if not any(condition(item))],
    )


def nested_dataclass(*args, **kwargs):
    def wrapper(cls):
        cls = dataclass(cls, **kwargs)
        original_init = cls.__init__

        def __init__(self, *args, **kwargs):
            for name, value in kwargs.items():
                field_type = cls.__annotations__.get(name, None)
                if is_dataclass(field_type) and isinstance(value, dict):
                    new_obj = field_type(**value)
                    kwargs[name] = new_obj
            original_init(self, *args, **kwargs)

        cls.__init__ = __init__
        return cls

    return wrapper(args[0]) if args else wrapper
